# Code generated by moonworm : https://github.com/bugout-dev/moonworm
# Moonworm version : 0.2.2

import argparse
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from brownie import Contract, network, project
from brownie.network.contract import ContractContainer
from eth_typing.evm import ChecksumAddress


PROJECT_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
BUILD_DIRECTORY = os.path.join(PROJECT_DIRECTORY, "build", "contracts")


def boolean_argument_type(raw_value: str) -> bool:
    TRUE_VALUES = ["1", "t", "y", "true", "yes"]
    FALSE_VALUES = ["0", "f", "n", "false", "no"]

    if raw_value.lower() in TRUE_VALUES:
        return True
    elif raw_value.lower() in FALSE_VALUES:
        return False

    raise ValueError(
        f"Invalid boolean argument: {raw_value}. Value must be one of: {','.join(TRUE_VALUES + FALSE_VALUES)}"
    )


def bytes_argument_type(raw_value: str) -> str:
    return raw_value


def get_abi_json(abi_name: str) -> List[Dict[str, Any]]:
    abi_full_path = os.path.join(BUILD_DIRECTORY, f"{abi_name}.json")
    if not os.path.isfile(abi_full_path):
        raise IOError(
            f"File does not exist: {abi_full_path}. Maybe you have to compile the smart contracts?"
        )

    with open(abi_full_path, "r") as ifp:
        build = json.load(ifp)

    abi_json = build.get("abi")
    if abi_json is None:
        raise ValueError(f"Could not find ABI definition in: {abi_full_path}")

    return abi_json


def contract_from_build(abi_name: str) -> ContractContainer:
    # This is workaround because brownie currently doesn't support loading the same project multiple
    # times. This causes problems when using multiple contracts from the same project in the same
    # python project.
    PROJECT = project.main.Project("moonworm", Path(PROJECT_DIRECTORY))

    abi_full_path = os.path.join(BUILD_DIRECTORY, f"{abi_name}.json")
    if not os.path.isfile(abi_full_path):
        raise IOError(
            f"File does not exist: {abi_full_path}. Maybe you have to compile the smart contracts?"
        )

    with open(abi_full_path, "r") as ifp:
        build = json.load(ifp)

    return ContractContainer(PROJECT, build)


class StatsFacet:
    def __init__(self, contract_address: Optional[ChecksumAddress]):
        self.contract_name = "StatsFacet"
        self.address = contract_address
        self.contract = None
        self.abi = get_abi_json("StatsFacet")
        if self.address is not None:
            self.contract: Optional[Contract] = Contract.from_abi(
                self.contract_name, self.address, self.abi
            )

    def deploy(self, transaction_config):
        contract_class = contract_from_build(self.contract_name)
        deployed_contract = contract_class.deploy(transaction_config)
        self.address = deployed_contract.address
        self.contract = deployed_contract
        return deployed_contract.tx

    def assert_contract_is_instantiated(self) -> None:
        if self.contract is None:
            raise Exception("contract has not been instantiated")

    def verify_contract(self):
        self.assert_contract_is_instantiated()
        contract_class = contract_from_build(self.contract_name)
        contract_class.publish_source(self.contract)

    def get_accuracy(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getAccuracy.call(_dna, block_identifier=block_number)

    def get_attack(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getAttack.call(_dna, block_identifier=block_number)

    def get_attack_speed(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getAttackSpeed.call(_dna, block_identifier=block_number)

    def get_defense(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getDefense.call(_dna, block_identifier=block_number)

    def get_endurance_score(
        self, token_id: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getEnduranceScore.call(
            token_id, block_identifier=block_number
        )

    def get_intelligence_score(
        self, token_id: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getIntelligenceScore.call(
            token_id, block_identifier=block_number
        )

    def get_magic(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getMagic.call(_dna, block_identifier=block_number)

    def get_movement_speed(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getMovementSpeed.call(_dna, block_identifier=block_number)

    def get_power_score(
        self, token_id: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getPowerScore.call(token_id, block_identifier=block_number)

    def get_resistance(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getResistance.call(_dna, block_identifier=block_number)

    def get_speed_score(
        self, token_id: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getSpeedScore.call(token_id, block_identifier=block_number)

    def get_stats(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getStats.call(_dna, block_identifier=block_number)

    def get_unicorn_body_parts(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getUnicornBodyParts.call(
            _dna, block_identifier=block_number
        )

    def get_unicorn_body_parts_local(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getUnicornBodyPartsLocal.call(
            _dna, block_identifier=block_number
        )

    def get_unicorn_metadata(
        self, _token_id: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getUnicornMetadata.call(
            _token_id, block_identifier=block_number
        )

    def get_vitality(
        self, _dna: int, block_number: Optional[Union[str, int]] = "latest"
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getVitality.call(_dna, block_identifier=block_number)


def get_transaction_config(args: argparse.Namespace) -> Dict[str, Any]:
    signer = network.accounts.load(args.sender, args.password)
    transaction_config: Dict[str, Any] = {"from": signer}
    if args.gas_price is not None:
        transaction_config["gas_price"] = args.gas_price
    if args.max_fee_per_gas is not None:
        transaction_config["max_fee"] = args.max_fee_per_gas
    if args.max_priority_fee_per_gas is not None:
        transaction_config["priority_fee"] = args.max_priority_fee_per_gas
    if args.confirmations is not None:
        transaction_config["required_confs"] = args.confirmations
    if args.nonce is not None:
        transaction_config["nonce"] = args.nonce
    return transaction_config


def add_default_arguments(parser: argparse.ArgumentParser, transact: bool) -> None:
    parser.add_argument(
        "--network", required=True, help="Name of brownie network to connect to"
    )
    parser.add_argument(
        "--address", required=False, help="Address of deployed contract to connect to"
    )
    if not transact:
        parser.add_argument(
            "--block-number",
            required=False,
            type=int,
            help="Call at the given block number, defaults to latest",
        )
        return
    parser.add_argument(
        "--sender", required=True, help="Path to keystore file for transaction sender"
    )
    parser.add_argument(
        "--password",
        required=False,
        help="Password to keystore file (if you do not provide it, you will be prompted for it)",
    )
    parser.add_argument(
        "--gas-price", default=None, help="Gas price at which to submit transaction"
    )
    parser.add_argument(
        "--max-fee-per-gas",
        default=None,
        help="Max fee per gas for EIP1559 transactions",
    )
    parser.add_argument(
        "--max-priority-fee-per-gas",
        default=None,
        help="Max priority fee per gas for EIP1559 transactions",
    )
    parser.add_argument(
        "--confirmations",
        type=int,
        default=None,
        help="Number of confirmations to await before considering a transaction completed",
    )
    parser.add_argument(
        "--nonce", type=int, default=None, help="Nonce for the transaction (optional)"
    )
    parser.add_argument(
        "--value", default=None, help="Value of the transaction in wei(optional)"
    )
    parser.add_argument("--verbose", action="store_true", help="Print verbose output")


def handle_deploy(args: argparse.Namespace) -> None:
    network.connect(args.network)
    transaction_config = get_transaction_config(args)
    contract = StatsFacet(None)
    result = contract.deploy(transaction_config=transaction_config)
    print(result)
    print(result.info())


def handle_verify_contract(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.verify_contract()
    print(result)


def handle_get_accuracy(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_accuracy(_dna=args.dna_arg, block_number=args.block_number)
    print(result)


def handle_get_attack(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_attack(_dna=args.dna_arg, block_number=args.block_number)
    print(result)


def handle_get_attack_speed(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_attack_speed(
        _dna=args.dna_arg, block_number=args.block_number
    )
    print(result)


def handle_get_defense(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_defense(_dna=args.dna_arg, block_number=args.block_number)
    print(result)


def handle_get_endurance_score(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_endurance_score(
        token_id=args.token_id, block_number=args.block_number
    )
    print(result)


def handle_get_intelligence_score(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_intelligence_score(
        token_id=args.token_id, block_number=args.block_number
    )
    print(result)


def handle_get_magic(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_magic(_dna=args.dna_arg, block_number=args.block_number)
    print(result)


def handle_get_movement_speed(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_movement_speed(
        _dna=args.dna_arg, block_number=args.block_number
    )
    print(result)


def handle_get_power_score(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_power_score(
        token_id=args.token_id, block_number=args.block_number
    )
    print(result)


def handle_get_resistance(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_resistance(_dna=args.dna_arg, block_number=args.block_number)
    print(result)


def handle_get_speed_score(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_speed_score(
        token_id=args.token_id, block_number=args.block_number
    )
    print(result)


def handle_get_stats(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_stats(_dna=args.dna_arg, block_number=args.block_number)
    print(result)


def handle_get_unicorn_body_parts(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_unicorn_body_parts(
        _dna=args.dna_arg, block_number=args.block_number
    )
    print(result)


def handle_get_unicorn_body_parts_local(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_unicorn_body_parts_local(
        _dna=args.dna_arg, block_number=args.block_number
    )
    print(result)


def handle_get_unicorn_metadata(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_unicorn_metadata(
        _token_id=args.token_id_arg, block_number=args.block_number
    )
    print(result)


def handle_get_vitality(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = StatsFacet(args.address)
    result = contract.get_vitality(_dna=args.dna_arg, block_number=args.block_number)
    print(result)


def generate_cli() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="CLI for StatsFacet")
    parser.set_defaults(func=lambda _: parser.print_help())
    subcommands = parser.add_subparsers()

    deploy_parser = subcommands.add_parser("deploy")
    add_default_arguments(deploy_parser, True)
    deploy_parser.set_defaults(func=handle_deploy)

    verify_contract_parser = subcommands.add_parser("verify-contract")
    add_default_arguments(verify_contract_parser, False)
    verify_contract_parser.set_defaults(func=handle_verify_contract)

    get_accuracy_parser = subcommands.add_parser("get-accuracy")
    add_default_arguments(get_accuracy_parser, False)
    get_accuracy_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_accuracy_parser.set_defaults(func=handle_get_accuracy)

    get_attack_parser = subcommands.add_parser("get-attack")
    add_default_arguments(get_attack_parser, False)
    get_attack_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_attack_parser.set_defaults(func=handle_get_attack)

    get_attack_speed_parser = subcommands.add_parser("get-attack-speed")
    add_default_arguments(get_attack_speed_parser, False)
    get_attack_speed_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_attack_speed_parser.set_defaults(func=handle_get_attack_speed)

    get_defense_parser = subcommands.add_parser("get-defense")
    add_default_arguments(get_defense_parser, False)
    get_defense_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_defense_parser.set_defaults(func=handle_get_defense)

    get_endurance_score_parser = subcommands.add_parser("get-endurance-score")
    add_default_arguments(get_endurance_score_parser, False)
    get_endurance_score_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    get_endurance_score_parser.set_defaults(func=handle_get_endurance_score)

    get_intelligence_score_parser = subcommands.add_parser("get-intelligence-score")
    add_default_arguments(get_intelligence_score_parser, False)
    get_intelligence_score_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    get_intelligence_score_parser.set_defaults(func=handle_get_intelligence_score)

    get_magic_parser = subcommands.add_parser("get-magic")
    add_default_arguments(get_magic_parser, False)
    get_magic_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_magic_parser.set_defaults(func=handle_get_magic)

    get_movement_speed_parser = subcommands.add_parser("get-movement-speed")
    add_default_arguments(get_movement_speed_parser, False)
    get_movement_speed_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_movement_speed_parser.set_defaults(func=handle_get_movement_speed)

    get_power_score_parser = subcommands.add_parser("get-power-score")
    add_default_arguments(get_power_score_parser, False)
    get_power_score_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    get_power_score_parser.set_defaults(func=handle_get_power_score)

    get_resistance_parser = subcommands.add_parser("get-resistance")
    add_default_arguments(get_resistance_parser, False)
    get_resistance_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_resistance_parser.set_defaults(func=handle_get_resistance)

    get_speed_score_parser = subcommands.add_parser("get-speed-score")
    add_default_arguments(get_speed_score_parser, False)
    get_speed_score_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    get_speed_score_parser.set_defaults(func=handle_get_speed_score)

    get_stats_parser = subcommands.add_parser("get-stats")
    add_default_arguments(get_stats_parser, False)
    get_stats_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_stats_parser.set_defaults(func=handle_get_stats)

    get_unicorn_body_parts_parser = subcommands.add_parser("get-unicorn-body-parts")
    add_default_arguments(get_unicorn_body_parts_parser, False)
    get_unicorn_body_parts_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_unicorn_body_parts_parser.set_defaults(func=handle_get_unicorn_body_parts)

    get_unicorn_body_parts_local_parser = subcommands.add_parser(
        "get-unicorn-body-parts-local"
    )
    add_default_arguments(get_unicorn_body_parts_local_parser, False)
    get_unicorn_body_parts_local_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_unicorn_body_parts_local_parser.set_defaults(
        func=handle_get_unicorn_body_parts_local
    )

    get_unicorn_metadata_parser = subcommands.add_parser("get-unicorn-metadata")
    add_default_arguments(get_unicorn_metadata_parser, False)
    get_unicorn_metadata_parser.add_argument(
        "--token-id-arg", required=True, help="Type: uint256", type=int
    )
    get_unicorn_metadata_parser.set_defaults(func=handle_get_unicorn_metadata)

    get_vitality_parser = subcommands.add_parser("get-vitality")
    add_default_arguments(get_vitality_parser, False)
    get_vitality_parser.add_argument(
        "--dna-arg", required=True, help="Type: uint256", type=int
    )
    get_vitality_parser.set_defaults(func=handle_get_vitality)

    return parser


def main() -> None:
    parser = generate_cli()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



import argparse
from concurrent.futures import as_completed, ThreadPoolExecutor, Future
import csv
import json
from time import sleep
import sys
from typing import Any, Dict, List, Optional, Set, Tuple

from brownie import network
from brownie.network import chain
from tqdm import tqdm

from . import MetadataFacet
from . import StatsFacet
from eth_typing.evm import ChecksumAddress


def unicorn_dnas(
    contract_address: ChecksumAddress,
    token_ids: List[int],
    block_number: Optional[int] = None,
    num_workers: int = 1,
    timeout: float = 30.0,
    checkpoint_file: Optional[str] = None,
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    if block_number is None:
        block_number = len(chain) - 1

    contract = MetadataFacet.MetadataFacet(contract_address)

    results: List[Dict[str, Any]] = []
    checkpointed_token_ids: Set[int] = set()
    if checkpoint_file is not None:
        with open(checkpoint_file, "r") as ifp:
            for line in ifp:
                stripped_line = line.strip()
                if stripped_line:
                    result = json.loads(stripped_line)
                    if result.get("dna") is not None:
                        results.append(result)
                        checkpointed_token_ids.add(result["token_id"])

    errors: List[Dict[str, Any]] = []

    submission_progress_bar = tqdm(
        total=len(token_ids) - len(checkpointed_token_ids),
        desc="Submitting requests for unicorn DNAs",
    )

    dna_progress_bar = tqdm(
        total=len(token_ids) - len(checkpointed_token_ids),
        desc="Retrieving unicorn DNAs",
    )

    jobs: Dict[Future, int] = {}
    with ThreadPoolExecutor(max_workers=num_workers) as executor:
        for token_id in token_ids:
            if token_id in checkpointed_token_ids:
                continue
            submission_progress_bar.update()
            future = executor.submit(contract.get_dna, token_id, block_number)
            jobs[future] = token_id

    for future in as_completed(jobs, timeout):
        dna_progress_bar.update()
        token_id = jobs[future]
        try:
            dna = future.result()
            result = {
                "token_id": token_id,
                "block_number": block_number,
                "dna": str(dna),
            }
            results.append(result)
        except Exception as e:
            error = {
                "token_id": token_id,
                "block_number": block_number,
                "error": f"Failed to retrieve DNA: {str(e)}",
            }
            errors.append(error)

    return results, errors


def unicorn_metadata(
    contract_address: ChecksumAddress,
    token_ids: List[int],
    block_number: Optional[int] = None,
    num_workers: int = 1,
    timeout: float = 30.0,
    checkpoint_file: Optional[str] = None,
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    if block_number is None:
        block_number = len(chain) - 1

    contract = StatsFacet.StatsFacet(contract_address)

    results: List[Dict[str, Any]] = []
    checkpointed_token_ids: Set[int] = set()
    if checkpoint_file is not None:
        with open(checkpoint_file, "r") as ifp:
            for line in ifp:
                stripped_line = line.strip()
                if stripped_line:
                    result = json.loads(stripped_line)
                    if (
                        result.get("class_number") is not None
                        and result.get("lifecycle_stage") is not None
                    ):
                        results.append(result)
                        checkpointed_token_ids.add(result["token_id"])

    errors: List[Dict[str, Any]] = []

    submission_progress_bar = tqdm(
        total=len(token_ids) - len(checkpointed_token_ids),
        desc="Submitting requests for unicorn classes",
    )

    metadata_progress_bar = tqdm(
        total=len(token_ids) - len(checkpointed_token_ids),
        desc="Retrieving unicorn classes",
    )

    jobs: Dict[Future, int] = {}
    with ThreadPoolExecutor(max_workers=num_workers) as executor:
        for token_id in token_ids:
            if token_id in checkpointed_token_ids:
                continue
            submission_progress_bar.update()
            future = executor.submit(
                contract.get_unicorn_metadata, token_id, block_number
            )
            jobs[future] = token_id

    for future in as_completed(jobs, timeout):
        metadata_progress_bar.update()
        token_id = jobs[future]
        try:
            metadata = future.result()
            result = {
                "token_id": token_id,
                "block_number": block_number,
                "lifecycle_stage": metadata[3],
                "class_number": metadata[-2],
            }
            results.append(result)
        except Exception as e:
            error = {
                "token_id": token_id,
                "block_number": block_number,
                "error": f"Failed to retrieve DNA: {str(e)}",
            }
            errors.append(error)

    return results, errors


def unicorn_mythic_body_parts(
    contract_address: ChecksumAddress,
    dnas: List[Dict[str, Any]],
    block_number: Optional[int] = None,
    num_workers: int = 1,
    timeout: float = 30.0,
    checkpoint_file: Optional[str] = None,
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    if block_number is None:
        block_number = len(chain) - 1

    results: List[Dict[str, Any]] = []
    errors: List[Dict[str, Any]] = []

    checkpointed_token_ids: Set[int] = set()
    if checkpoint_file is not None:
        with open(checkpoint_file, "r") as ifp:
            for line in ifp:
                stripped_line = line.strip()
                if stripped_line:
                    result = json.loads(stripped_line)
                    if result.get("num_mythic_body_parts") is not None:
                        results.append(result)
                        checkpointed_token_ids.add(result["token_id"])

    submission_progress_bar = tqdm(
        total=len(dnas) - len(checkpointed_token_ids),
        desc="Submitting requests for number of unicorn mythic body parts",
    )

    mythic_progress_bar = tqdm(
        total=len(dnas) - len(checkpointed_token_ids),
        desc="Retrieving number of unicorn mythic body parts",
    )

    contract = StatsFacet.StatsFacet(contract_address)

    jobs: Dict[Future, int] = {}
    with ThreadPoolExecutor(max_workers=num_workers) as executor:
        for item in dnas:
            if item["token_id"] in checkpointed_token_ids:
                continue
            submission_progress_bar.update()
            future = executor.submit(
                contract.get_unicorn_body_parts,
                item["dna"],
                item["block_number"],
            )
            jobs[future] = item

    for future in as_completed(jobs, timeout):
        mythic_progress_bar.update()
        item = jobs[future]
        try:
            body_parts = future.result()
            result = {
                **item,
                "num_mythic_body_parts": body_parts[-1],
            }
            results.append(result)
        except Exception as e:
            error = {
                **item,
                "error": f"Failed to retrieve DNA: {str(e)}",
            }
            errors.append(error)

    return results, errors


def handle_dnas(args: argparse.Namespace) -> None:
    network.connect(args.network)
    if args.end is None:
        args.end = args.start
    assert args.start <= args.end, "Starting token ID must not exceed ending token ID"
    token_ids = range(args.start, args.end + 1)
    results, errors = unicorn_dnas(
        args.address,
        token_ids,
        args.block_number,
        args.num_workers,
        args.timeout,
        args.checkpoint,
    )

    for result in results:
        print(json.dumps(result))

    if args.update_checkpoint and args.checkpoint is not None:
        with open(args.checkpoint, "w") as ofp:
            for result in results:
                print(json.dumps(result), file=ofp)

    for error in errors:
        print(json.dumps(error), file=sys.stderr)


def handle_metadata(args: argparse.Namespace) -> None:
    network.connect(args.network)
    if args.end is None:
        args.end = args.start
    assert args.start <= args.end, "Starting token ID must not exceed ending token ID"
    token_ids = range(args.start, args.end + 1)
    results, errors = unicorn_metadata(
        args.address,
        token_ids,
        args.block_number,
        args.num_workers,
        args.timeout,
        args.checkpoint,
    )

    for result in results:
        print(json.dumps(result))

    if args.update_checkpoint and args.checkpoint is not None:
        with open(args.checkpoint, "w") as ofp:
            for result in results:
                print(json.dumps(result), file=ofp)

    for error in errors:
        print(json.dumps(error), file=sys.stderr)


def handle_mythic_body_parts(args: argparse.Namespace) -> None:
    network.connect(args.network)

    dnas: List[Dict[str, Any]] = []
    with open(args.dnas, "r") as ifp:
        for line in ifp:
            dnas.append(json.loads(line.strip()))

    results, errors = unicorn_mythic_body_parts(
        args.address,
        dnas,
        args.block_number,
        args.num_workers,
        args.timeout,
        args.checkpoint,
    )

    for result in results:
        json.dump(result, fp=sys.stdout)
        print("")

    if args.update_checkpoint and args.checkpoint is not None:
        with open(args.checkpoint, "w") as ofp:
            for result in results:
                print(json.dumps(result), file=ofp)

    for error in errors:
        json.dump(error, fp=sys.stderr)
        print("", file=sys.stderr)


def handle_merge(args: argparse.Namespace) -> None:
    metadata_index: Dict[int, Dict[str, Any]] = {}
    mythic_body_parts_index: Dict[int, Dict[str, Any]] = {}

    with open(args.metadata, "r") as metadata_ifp, open(
        args.mythic_body_parts, "r"
    ) as mythic_body_parts_ifp:
        for line in metadata_ifp:
            item = json.loads(line.strip())
            metadata_index[item["token_id"]] = item

        for line in mythic_body_parts_ifp:
            item = json.loads(line.strip())
            mythic_body_parts_index[item["token_id"]] = item

    hidden_classes = {1, 5, 8}
    for token_id, data in metadata_index.items():
        if token_id not in mythic_body_parts_index:
            print(
                f"Token ID in metadata but not in mythic-body-parts: {token_id}",
                file=sys.stderr,
            )
        else:
            mythic_body_parts_data = mythic_body_parts_index[token_id]
            result = {**data, **mythic_body_parts_data}
            del result["block_number"]
            result["metadata_block_number"] = data["block_number"]
            result["mythic_body_parts_block_number"] = mythic_body_parts_data[
                "block_number"
            ]
            if result["lifecycle_stage"] == 0:
                result["num_mythic_body_parts"] = 0
            result["is_mythic"] = result["num_mythic_body_parts"] > 0
            result["is_hidden_class"] = result["class_number"] in hidden_classes
            print(json.dumps(result), file=sys.stdout)


def handle_sob(args: argparse.Namespace) -> None:
    token_metadata_index: Dict[str, Dict[str, Any]] = {}
    with open(args.merged, "r") as ifp:
        for line in ifp:
            item = json.loads(line.strip())
            token_metadata_index[str(item["token_id"])] = item

    moonstream_data: List[Dict[str, Any]] = []
    with open(args.moonstream, "r") as ifp:
        reader = csv.DictReader(ifp)
        for row in reader:
            moonstream_data.append(row)

    breeding_events: List[Dict[str, Any]] = []
    hatching_events: List[Dict[str, Any]] = []

    for event in moonstream_data:
        if event["event_type"] == "breeding":
            event["milestone_1"] = 50
            breeding_events.append(event)
        elif event["event_type"] == "hatchingEggs":
            metadata = token_metadata_index[event["token"]]
            event["milestone_1"] = 0
            if metadata["is_mythic"]:
                event["milestone_1"] = 20
            hatching_events.append(event)
        else:
            # Other conditions in this if statement in the future.
            pass

    player_points: Dict[str, Dict[str, int]] = {}
    for event in breeding_events:
        player = event["player_wallet"]
        if player_points.get(player) is None:
            player_points[player] = {
                "milestone_1": 0,
                "num_breeds": 0,
                "num_hatches": 0,
                "num_mythic_hatches": 0,
            }
        player_points[player]["milestone_1"] += event["milestone_1"]
        player_points[player]["num_breeds"] += 1
    for event in hatching_events:
        player = event["player_wallet"]
        if player_points.get(player) is None:
            player_points[player] = {
                "milestone_1": 0,
                "num_breeds": 0,
                "num_hatches": 0,
                "num_mythic_hatches": 0,
            }
        player_points[player]["milestone_1"] += event["milestone_1"]
        player_points[player]["num_hatches"] += 1
        player_points[player]["num_mythic_hatches"] += int(event["milestone_1"] > 0)

    scores: List[Dict[str, Any]] = []
    for player, points in player_points.items():
        scores.append(
            {
                "address": player,
                "score": points["milestone_1"],
                "points_data": points,
            }
        )

    scores.sort(key=lambda item: item["score"], reverse=True)

    print(json.dumps(scores))


def generate_cli() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Crypto Unicorns genetics crawler")
    subparsers = parser.add_subparsers()

    dnas_parser = subparsers.add_parser("dnas")
    StatsFacet.add_default_arguments(dnas_parser, False)
    dnas_parser.add_argument(
        "--start",
        type=int,
        required=True,
        help="Starting token ID to get DNA for.",
    )
    dnas_parser.add_argument(
        "--end",
        type=int,
        required=False,
        help="Ending token ID to get DNA for. (If not set, just gets the DNA for the token with the --start token ID.)",
    )
    dnas_parser.add_argument(
        "-j",
        "--num-workers",
        type=int,
        default=1,
        help="Maximum number of concurrent threads to use when crawling",
    )
    dnas_parser.add_argument(
        "-t",
        "--timeout",
        type=float,
        default=30.0,
        help="Number of seconds to wait for each crawl response",
    )
    dnas_parser.add_argument(
        "--checkpoint",
        default=None,
        help="Checkpoint file",
    )
    dnas_parser.add_argument(
        "-u",
        "--update-checkpoint",
        action="store_true",
        help="If you have set a checkpoint, this updates the checkpoint file in place",
    )

    dnas_parser.set_defaults(func=handle_dnas)

    metadata_parser = subparsers.add_parser("metadata")
    StatsFacet.add_default_arguments(metadata_parser, False)
    metadata_parser.add_argument(
        "--start",
        type=int,
        required=True,
        help="Starting token ID to get DNA for.",
    )
    metadata_parser.add_argument(
        "--end",
        type=int,
        required=False,
        help="Ending token ID to get DNA for. (If not set, just gets the DNA for the token with the --start token ID.)",
    )
    metadata_parser.add_argument(
        "-j",
        "--num-workers",
        type=int,
        default=1,
        help="Maximum number of concurrent threads to use when crawling",
    )
    metadata_parser.add_argument(
        "-t",
        "--timeout",
        type=float,
        default=30.0,
        help="Number of seconds to wait for each crawl response",
    )
    metadata_parser.add_argument(
        "--checkpoint",
        default=None,
        help="Checkpoint file",
    )
    metadata_parser.add_argument(
        "-u",
        "--update-checkpoint",
        action="store_true",
        help="If you have set a checkpoint, this updates the checkpoint file in place",
    )

    metadata_parser.set_defaults(func=handle_metadata)

    mythic_body_parts_parser = subparsers.add_parser("mythic-body-parts")
    StatsFacet.add_default_arguments(mythic_body_parts_parser, False)
    mythic_body_parts_parser.add_argument(
        "--dnas",
        required=True,
        help='Path to JSON file containing results of "autocorns biologist dnas".',
    )
    mythic_body_parts_parser.add_argument(
        "-j",
        "--num-workers",
        type=int,
        default=1,
        help="Maximum number of concurrent threads to use when crawling",
    )
    mythic_body_parts_parser.add_argument(
        "-t",
        "--timeout",
        type=float,
        default=30.0,
        help="Number of seconds to wait for each crawl response",
    )
    mythic_body_parts_parser.add_argument(
        "--checkpoint",
        default=None,
        help="Checkpoint file",
    )
    mythic_body_parts_parser.add_argument(
        "-u",
        "--update-checkpoint",
        action="store_true",
        help="If you have set a checkpoint, this updates the checkpoint file in place",
    )

    mythic_body_parts_parser.set_defaults(func=handle_mythic_body_parts)

    merge_parser = subparsers.add_parser("merge")
    merge_parser.add_argument(
        "--metadata",
        required=True,
        help='Metadata file generated by "autocorns biologist metadata"',
    )
    merge_parser.add_argument(
        "--mythic-body-parts",
        required=True,
        help='Mythic body parts file generated by "autocorns biologist mythic-body-parts"',
    )

    merge_parser.set_defaults(func=handle_merge)

    sob_parser = subparsers.add_parser("sob")
    sob_parser.add_argument(
        "--merged",
        required=True,
        help='Merged file generated by "autocorns biologist merge"',
    )
    sob_parser.add_argument(
        "--moonstream",
        required=True,
        help="File provided by Moonstream",
    )

    sob_parser.set_defaults(func=handle_sob)

    return parser


if __name__ == "__main__":
    parser = generate_cli()
    args = parser.parse_args()
    args.func(args)



# Code generated by moonworm : https://github.com/bugout-dev/moonworm
# Moonworm version : 0.1.15

import argparse
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from brownie import Contract, network, project
from brownie.network.contract import ContractContainer
from eth_typing.evm import ChecksumAddress


PROJECT_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
BUILD_DIRECTORY = os.path.join(PROJECT_DIRECTORY, "build", "contracts")


def boolean_argument_type(raw_value: str) -> bool:
    TRUE_VALUES = ["1", "t", "y", "true", "yes"]
    FALSE_VALUES = ["0", "f", "n", "false", "no"]

    if raw_value.lower() in TRUE_VALUES:
        return True
    elif raw_value.lower() in FALSE_VALUES:
        return False

    raise ValueError(
        f"Invalid boolean argument: {raw_value}. Value must be one of: {','.join(TRUE_VALUES + FALSE_VALUES)}"
    )


def bytes_argument_type(raw_value: str) -> bytes:
    return raw_value.encode()


def get_abi_json(abi_name: str) -> List[Dict[str, Any]]:
    abi_full_path = os.path.join(BUILD_DIRECTORY, f"{abi_name}.json")
    if not os.path.isfile(abi_full_path):
        raise IOError(
            f"File does not exist: {abi_full_path}. Maybe you have to compile the smart contracts?"
        )

    with open(abi_full_path, "r") as ifp:
        build = json.load(ifp)

    abi_json = build.get("abi")
    if abi_json is None:
        raise ValueError(f"Could not find ABI definition in: {abi_full_path}")

    return abi_json


def contract_from_build(abi_name: str) -> ContractContainer:
    # This is workaround because brownie currently doesn't support loading the same project multiple
    # times. This causes problems when using multiple contracts from the same project in the same
    # python project.
    PROJECT = project.main.Project("moonworm", Path(PROJECT_DIRECTORY))

    abi_full_path = os.path.join(BUILD_DIRECTORY, f"{abi_name}.json")
    if not os.path.isfile(abi_full_path):
        raise IOError(
            f"File does not exist: {abi_full_path}. Maybe you have to compile the smart contracts?"
        )

    with open(abi_full_path, "r") as ifp:
        build = json.load(ifp)

    return ContractContainer(PROJECT, build)


class DarkForest:
    def __init__(self, contract_address: Optional[ChecksumAddress]):
        self.contract_name = "DarkForest"
        self.address = contract_address
        self.contract = None
        self.abi = get_abi_json("DarkForest")
        if self.address is not None:
            self.contract: Optional[Contract] = Contract.from_abi(
                self.contract_name, self.address, self.abi
            )

    def deploy(
        self,
        _crypto_unicorns_address: ChecksumAddress,
        _unicorn_milk_token_address: ChecksumAddress,
        _terminus_address: ChecksumAddress,
        _unicorn_milk_staking_reward: int,
        _stake_period_seconds: int,
        _vrf_coordinator_address: ChecksumAddress,
        _link_token_address: ChecksumAddress,
        _chainlink_vrf_fee: int,
        _chainlink_vrf_keyhash: bytes,
        transaction_config,
    ):
        contract_class = contract_from_build(self.contract_name)
        deployed_contract = contract_class.deploy(
            _crypto_unicorns_address,
            _unicorn_milk_token_address,
            _terminus_address,
            _unicorn_milk_staking_reward,
            _stake_period_seconds,
            _vrf_coordinator_address,
            _link_token_address,
            _chainlink_vrf_fee,
            _chainlink_vrf_keyhash,
            transaction_config,
        )
        self.address = deployed_contract.address
        self.contract = deployed_contract

    def assert_contract_is_instantiated(self) -> None:
        if self.contract is None:
            raise Exception("contract has not been instantiated")

    def verify_contract(self):
        self.assert_contract_is_instantiated()
        contract_class = contract_from_build(self.contract_name)
        contract_class.publish_source(self.contract)

    def chainlink_vrf_fee(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.chainlinkVRFFee.call()

    def chainlink_vrf_keyhash(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.chainlinkVRFKeyhash.call()

    def complete_sacrifice(self, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.completeSacrifice(transaction_config)

    def contract_is_initialized(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.contractIsInitialized.call()

    def crypto_unicorns_address(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.cryptoUnicornsAddress.call()

    def current_sacrifice_id(self, sacrificer: ChecksumAddress) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.currentSacrificeId.call(sacrificer)

    def drain_erc20(self, token_address: ChecksumAddress, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.drainERC20(token_address, transaction_config)

    def exit_forest(self, token_id: int, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.exitForest(token_id, transaction_config)

    def lootbox_pool_ids(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.lootboxPoolIds.call()

    def num_staked(self, _staker: ChecksumAddress) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.numStaked.call(_staker)

    def offer_sacrifice(self, amount: int, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.offerSacrifice(amount, transaction_config)

    def on_erc721_received(
        self,
        operator: ChecksumAddress,
        from_: ChecksumAddress,
        token_id: int,
        data: bytes,
        transaction_config,
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.onERC721Received(
            operator, from_, token_id, data, transaction_config
        )

    def owner(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.owner.call()

    def raw_fulfill_randomness(
        self, request_id: bytes, randomness: int, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.rawFulfillRandomness(
            request_id, randomness, transaction_config
        )

    def remaining_shadowcorn_eggs(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.remainingShadowcornEggs.call()

    def renounce_ownership(self, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.renounceOwnership(transaction_config)

    def rescue_unicorn(self, token_id: int, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.rescueUnicorn(token_id, transaction_config)

    def reset_sacrifice_for_sacrificer(
        self, sacrificer: ChecksumAddress, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.resetSacrificeForSacrificer(sacrificer, transaction_config)

    def set_chainlink_vrf_fee(self, _fee: int, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setChainlinkVRFFee(_fee, transaction_config)

    def set_chainlink_vrf_keyhash(self, _keyhash: bytes, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setChainlinkVRFKeyhash(_keyhash, transaction_config)

    def set_crypto_unicorns_address(
        self, _crypto_unicorns_address: ChecksumAddress, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setCryptoUnicornsAddress(
            _crypto_unicorns_address, transaction_config
        )

    def set_lootbox_pool_ids(self, _lootbox_pool_ids: List, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setLootboxPoolIds(_lootbox_pool_ids, transaction_config)

    def set_shadowcorn_egg_bonuses_to_unicorn_milk(
        self, _bonuses: List, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setShadowcornEggBonusesToUnicornMilk(
            _bonuses, transaction_config
        )

    def set_shadowcorn_pool_ids(
        self, _shadowcorn_pool_ids: List, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setShadowcornPoolIds(
            _shadowcorn_pool_ids, transaction_config
        )

    def set_stake_period_seconds(
        self, _stake_period_seconds: int, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setStakePeriodSeconds(
            _stake_period_seconds, transaction_config
        )

    def set_terminus_address(
        self, _terminus_address: ChecksumAddress, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setTerminusAddress(_terminus_address, transaction_config)

    def set_unicorn_milk_staking_reward(
        self, _unicorn_milk_staking_reward: int, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setUnicornMilkStakingReward(
            _unicorn_milk_staking_reward, transaction_config
        )

    def set_unicorn_milk_thresholds(
        self, _unicorn_milk_thresholds: List, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setUnicornMilkThresholds(
            _unicorn_milk_thresholds, transaction_config
        )

    def set_unicorn_milk_token_address(
        self, _unicorn_milk_token_address: ChecksumAddress, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setUnicornMilkTokenAddress(
            _unicorn_milk_token_address, transaction_config
        )

    def shadowcorn_egg_bonuses_to_unicorn_milk(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.shadowcornEggBonusesToUnicornMilk.call()

    def shadowcorn_pool_ids(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.shadowcornPoolIds.call()

    def stake_period_seconds(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.stakePeriodSeconds.call()

    def staker(self, token_id: int) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.staker.call(token_id)

    def surrender_terminus_pools(self, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.surrenderTerminusPools(transaction_config)

    def terminus_address(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.terminusAddress.call()

    def token_of_staker_by_index(self, _staker: ChecksumAddress, index: int) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.tokenOfStakerByIndex.call(_staker, index)

    def transfer_ownership(self, new_owner: ChecksumAddress, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.transferOwnership(new_owner, transaction_config)

    def unicorn_milk_staking_reward(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.unicornMilkStakingReward.call()

    def unicorn_milk_thresholds(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.unicornMilkThresholds.call()

    def unicorn_milk_token_address(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.unicornMilkTokenAddress.call()

    def unstakes_at(self, token_id: int) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.unstakesAt.call(token_id)

    def view_sacrifice(self, _request_id: bytes) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.viewSacrifice.call(_request_id)

    def withdraw_erc20(
        self, token_address: ChecksumAddress, amount: int, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.withdrawERC20(token_address, amount, transaction_config)


def get_transaction_config(args: argparse.Namespace) -> Dict[str, Any]:
    signer = network.accounts.load(args.sender, args.password)
    transaction_config: Dict[str, Any] = {"from": signer}
    if args.gas_price is not None:
        transaction_config["gas_price"] = args.gas_price
    if args.max_fee_per_gas is not None:
        transaction_config["max_fee"] = args.max_fee_per_gas
    if args.max_priority_fee_per_gas is not None:
        transaction_config["priority_fee"] = args.max_priority_fee_per_gas
    if args.confirmations is not None:
        transaction_config["required_confs"] = args.confirmations
    if args.nonce is not None:
        transaction_config["nonce"] = args.nonce
    return transaction_config


def add_default_arguments(parser: argparse.ArgumentParser, transact: bool) -> None:
    parser.add_argument(
        "--network", required=True, help="Name of brownie network to connect to"
    )
    parser.add_argument(
        "--address", required=False, help="Address of deployed contract to connect to"
    )
    if not transact:
        return
    parser.add_argument(
        "--sender", required=True, help="Path to keystore file for transaction sender"
    )
    parser.add_argument(
        "--password",
        required=False,
        help="Password to keystore file (if you do not provide it, you will be prompted for it)",
    )
    parser.add_argument(
        "--gas-price", default=None, help="Gas price at which to submit transaction"
    )
    parser.add_argument(
        "--max-fee-per-gas",
        default=None,
        help="Max fee per gas for EIP1559 transactions",
    )
    parser.add_argument(
        "--max-priority-fee-per-gas",
        default=None,
        help="Max priority fee per gas for EIP1559 transactions",
    )
    parser.add_argument(
        "--confirmations",
        type=int,
        default=None,
        help="Number of confirmations to await before considering a transaction completed",
    )
    parser.add_argument(
        "--nonce", type=int, default=None, help="Nonce for the transaction (optional)"
    )
    parser.add_argument(
        "--value", default=None, help="Value of the transaction in wei(optional)"
    )


def handle_deploy(args: argparse.Namespace) -> None:
    network.connect(args.network)
    transaction_config = get_transaction_config(args)
    contract = DarkForest(None)
    result = contract.deploy(
        _crypto_unicorns_address=args.crypto_unicorns_address_arg,
        _unicorn_milk_token_address=args.unicorn_milk_token_address_arg,
        _terminus_address=args.terminus_address_arg,
        _unicorn_milk_staking_reward=args.unicorn_milk_staking_reward_arg,
        _stake_period_seconds=args.stake_period_seconds_arg,
        _vrf_coordinator_address=args.vrf_coordinator_address_arg,
        _link_token_address=args.link_token_address_arg,
        _chainlink_vrf_fee=args.chainlink_vrf_fee_arg,
        _chainlink_vrf_keyhash=args.chainlink_vrf_keyhash_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_verify_contract(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.verify_contract()
    print(result)


def handle_chainlink_vrf_fee(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.chainlink_vrf_fee()
    print(result)


def handle_chainlink_vrf_keyhash(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.chainlink_vrf_keyhash()
    print(result)


def handle_complete_sacrifice(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.complete_sacrifice(transaction_config=transaction_config)
    print(result)


def handle_contract_is_initialized(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.contract_is_initialized()
    print(result)


def handle_crypto_unicorns_address(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.crypto_unicorns_address()
    print(result)


def handle_current_sacrifice_id(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.current_sacrifice_id(sacrificer=args.sacrificer)
    print(result)


def handle_drain_erc20(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.drain_erc20(
        token_address=args.token_address, transaction_config=transaction_config
    )
    print(result)


def handle_exit_forest(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.exit_forest(
        token_id=args.token_id, transaction_config=transaction_config
    )
    print(result)


def handle_lootbox_pool_ids(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.lootbox_pool_ids()
    print(result)


def handle_num_staked(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.num_staked(_staker=args.staker_arg)
    print(result)


def handle_offer_sacrifice(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.offer_sacrifice(
        amount=args.amount, transaction_config=transaction_config
    )
    print(result)


def handle_on_erc721_received(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.on_erc721_received(
        operator=args.operator,
        from_=args.from_arg,
        token_id=args.token_id,
        data=args.data,
        transaction_config=transaction_config,
    )
    print(result)


def handle_owner(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.owner()
    print(result)


def handle_raw_fulfill_randomness(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.raw_fulfill_randomness(
        request_id=args.request_id,
        randomness=args.randomness,
        transaction_config=transaction_config,
    )
    print(result)


def handle_remaining_shadowcorn_eggs(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.remaining_shadowcorn_eggs()
    print(result)


def handle_renounce_ownership(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.renounce_ownership(transaction_config=transaction_config)
    print(result)


def handle_rescue_unicorn(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.rescue_unicorn(
        token_id=args.token_id, transaction_config=transaction_config
    )
    print(result)


def handle_reset_sacrifice_for_sacrificer(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.reset_sacrifice_for_sacrificer(
        sacrificer=args.sacrificer, transaction_config=transaction_config
    )
    print(result)


def handle_set_chainlink_vrf_fee(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_chainlink_vrf_fee(
        _fee=args.fee_arg, transaction_config=transaction_config
    )
    print(result)


def handle_set_chainlink_vrf_keyhash(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_chainlink_vrf_keyhash(
        _keyhash=args.keyhash_arg, transaction_config=transaction_config
    )
    print(result)


def handle_set_crypto_unicorns_address(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_crypto_unicorns_address(
        _crypto_unicorns_address=args.crypto_unicorns_address_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_lootbox_pool_ids(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_lootbox_pool_ids(
        _lootbox_pool_ids=args.lootbox_pool_ids_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_shadowcorn_egg_bonuses_to_unicorn_milk(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_shadowcorn_egg_bonuses_to_unicorn_milk(
        _bonuses=args.bonuses_arg, transaction_config=transaction_config
    )
    print(result)


def handle_set_shadowcorn_pool_ids(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_shadowcorn_pool_ids(
        _shadowcorn_pool_ids=args.shadowcorn_pool_ids_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_stake_period_seconds(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_stake_period_seconds(
        _stake_period_seconds=args.stake_period_seconds_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_terminus_address(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_terminus_address(
        _terminus_address=args.terminus_address_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_unicorn_milk_staking_reward(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_unicorn_milk_staking_reward(
        _unicorn_milk_staking_reward=args.unicorn_milk_staking_reward_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_unicorn_milk_thresholds(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_unicorn_milk_thresholds(
        _unicorn_milk_thresholds=args.unicorn_milk_thresholds_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_unicorn_milk_token_address(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_unicorn_milk_token_address(
        _unicorn_milk_token_address=args.unicorn_milk_token_address_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_shadowcorn_egg_bonuses_to_unicorn_milk(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.shadowcorn_egg_bonuses_to_unicorn_milk()
    print(result)


def handle_shadowcorn_pool_ids(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.shadowcorn_pool_ids()
    print(result)


def handle_stake_period_seconds(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.stake_period_seconds()
    print(result)


def handle_staker(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.staker(token_id=args.token_id)
    print(result)


def handle_surrender_terminus_pools(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.surrender_terminus_pools(transaction_config=transaction_config)
    print(result)


def handle_terminus_address(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.terminus_address()
    print(result)


def handle_token_of_staker_by_index(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.token_of_staker_by_index(
        _staker=args.staker_arg, index=args.index
    )
    print(result)


def handle_transfer_ownership(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.transfer_ownership(
        new_owner=args.new_owner, transaction_config=transaction_config
    )
    print(result)


def handle_unicorn_milk_staking_reward(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.unicorn_milk_staking_reward()
    print(result)


def handle_unicorn_milk_thresholds(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.unicorn_milk_thresholds()
    print(result)


def handle_unicorn_milk_token_address(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.unicorn_milk_token_address()
    print(result)


def handle_unstakes_at(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.unstakes_at(token_id=args.token_id)
    print(result)


def handle_view_sacrifice(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    result = contract.view_sacrifice(_request_id=args.request_id_arg)
    print(result)


def handle_withdraw_erc20(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = DarkForest(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.withdraw_erc20(
        token_address=args.token_address,
        amount=args.amount,
        transaction_config=transaction_config,
    )
    print(result)


def generate_cli() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="CLI for DarkForest")
    parser.set_defaults(func=lambda _: parser.print_help())
    subcommands = parser.add_subparsers()

    deploy_parser = subcommands.add_parser("deploy")
    add_default_arguments(deploy_parser, True)
    deploy_parser.add_argument(
        "--crypto-unicorns-address-arg", required=True, help="Type: address"
    )
    deploy_parser.add_argument(
        "--unicorn-milk-token-address-arg", required=True, help="Type: address"
    )
    deploy_parser.add_argument(
        "--terminus-address-arg", required=True, help="Type: address"
    )
    deploy_parser.add_argument(
        "--unicorn-milk-staking-reward-arg",
        required=True,
        help="Type: uint256",
        type=int,
    )
    deploy_parser.add_argument(
        "--stake-period-seconds-arg", required=True, help="Type: uint256", type=int
    )
    deploy_parser.add_argument(
        "--vrf-coordinator-address-arg", required=True, help="Type: address"
    )
    deploy_parser.add_argument(
        "--link-token-address-arg", required=True, help="Type: address"
    )
    deploy_parser.add_argument(
        "--chainlink-vrf-fee-arg", required=True, help="Type: uint256", type=int
    )
    deploy_parser.add_argument(
        "--chainlink-vrf-keyhash-arg",
        required=True,
        help="Type: bytes32",
        type=bytes_argument_type,
    )
    deploy_parser.set_defaults(func=handle_deploy)

    verify_contract_parser = subcommands.add_parser("verify-contract")
    add_default_arguments(verify_contract_parser, False)
    verify_contract_parser.set_defaults(func=handle_verify_contract)

    chainlink_vrf_fee_parser = subcommands.add_parser("chainlink-vrf-fee")
    add_default_arguments(chainlink_vrf_fee_parser, False)
    chainlink_vrf_fee_parser.set_defaults(func=handle_chainlink_vrf_fee)

    chainlink_vrf_keyhash_parser = subcommands.add_parser("chainlink-vrf-keyhash")
    add_default_arguments(chainlink_vrf_keyhash_parser, False)
    chainlink_vrf_keyhash_parser.set_defaults(func=handle_chainlink_vrf_keyhash)

    complete_sacrifice_parser = subcommands.add_parser("complete-sacrifice")
    add_default_arguments(complete_sacrifice_parser, True)
    complete_sacrifice_parser.set_defaults(func=handle_complete_sacrifice)

    contract_is_initialized_parser = subcommands.add_parser("contract-is-initialized")
    add_default_arguments(contract_is_initialized_parser, False)
    contract_is_initialized_parser.set_defaults(func=handle_contract_is_initialized)

    crypto_unicorns_address_parser = subcommands.add_parser("crypto-unicorns-address")
    add_default_arguments(crypto_unicorns_address_parser, False)
    crypto_unicorns_address_parser.set_defaults(func=handle_crypto_unicorns_address)

    current_sacrifice_id_parser = subcommands.add_parser("current-sacrifice-id")
    add_default_arguments(current_sacrifice_id_parser, False)
    current_sacrifice_id_parser.add_argument(
        "--sacrificer", required=True, help="Type: address"
    )
    current_sacrifice_id_parser.set_defaults(func=handle_current_sacrifice_id)

    drain_erc20_parser = subcommands.add_parser("drain-erc20")
    add_default_arguments(drain_erc20_parser, True)
    drain_erc20_parser.add_argument(
        "--token-address", required=True, help="Type: address"
    )
    drain_erc20_parser.set_defaults(func=handle_drain_erc20)

    exit_forest_parser = subcommands.add_parser("exit-forest")
    add_default_arguments(exit_forest_parser, True)
    exit_forest_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    exit_forest_parser.set_defaults(func=handle_exit_forest)

    lootbox_pool_ids_parser = subcommands.add_parser("lootbox-pool-ids")
    add_default_arguments(lootbox_pool_ids_parser, False)
    lootbox_pool_ids_parser.set_defaults(func=handle_lootbox_pool_ids)

    num_staked_parser = subcommands.add_parser("num-staked")
    add_default_arguments(num_staked_parser, False)
    num_staked_parser.add_argument("--staker-arg", required=True, help="Type: address")
    num_staked_parser.set_defaults(func=handle_num_staked)

    offer_sacrifice_parser = subcommands.add_parser("offer-sacrifice")
    add_default_arguments(offer_sacrifice_parser, True)
    offer_sacrifice_parser.add_argument(
        "--amount", required=True, help="Type: uint256", type=int
    )
    offer_sacrifice_parser.set_defaults(func=handle_offer_sacrifice)

    on_erc721_received_parser = subcommands.add_parser("on-erc721-received")
    add_default_arguments(on_erc721_received_parser, True)
    on_erc721_received_parser.add_argument(
        "--operator", required=True, help="Type: address"
    )
    on_erc721_received_parser.add_argument(
        "--from-arg", required=True, help="Type: address"
    )
    on_erc721_received_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    on_erc721_received_parser.add_argument(
        "--data", required=True, help="Type: bytes", type=bytes_argument_type
    )
    on_erc721_received_parser.set_defaults(func=handle_on_erc721_received)

    owner_parser = subcommands.add_parser("owner")
    add_default_arguments(owner_parser, False)
    owner_parser.set_defaults(func=handle_owner)

    raw_fulfill_randomness_parser = subcommands.add_parser("raw-fulfill-randomness")
    add_default_arguments(raw_fulfill_randomness_parser, True)
    raw_fulfill_randomness_parser.add_argument(
        "--request-id", required=True, help="Type: bytes32", type=bytes_argument_type
    )
    raw_fulfill_randomness_parser.add_argument(
        "--randomness", required=True, help="Type: uint256", type=int
    )
    raw_fulfill_randomness_parser.set_defaults(func=handle_raw_fulfill_randomness)

    remaining_shadowcorn_eggs_parser = subcommands.add_parser(
        "remaining-shadowcorn-eggs"
    )
    add_default_arguments(remaining_shadowcorn_eggs_parser, False)
    remaining_shadowcorn_eggs_parser.set_defaults(func=handle_remaining_shadowcorn_eggs)

    renounce_ownership_parser = subcommands.add_parser("renounce-ownership")
    add_default_arguments(renounce_ownership_parser, True)
    renounce_ownership_parser.set_defaults(func=handle_renounce_ownership)

    rescue_unicorn_parser = subcommands.add_parser("rescue-unicorn")
    add_default_arguments(rescue_unicorn_parser, True)
    rescue_unicorn_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    rescue_unicorn_parser.set_defaults(func=handle_rescue_unicorn)

    reset_sacrifice_for_sacrificer_parser = subcommands.add_parser(
        "reset-sacrifice-for-sacrificer"
    )
    add_default_arguments(reset_sacrifice_for_sacrificer_parser, True)
    reset_sacrifice_for_sacrificer_parser.add_argument(
        "--sacrificer", required=True, help="Type: address"
    )
    reset_sacrifice_for_sacrificer_parser.set_defaults(
        func=handle_reset_sacrifice_for_sacrificer
    )

    set_chainlink_vrf_fee_parser = subcommands.add_parser("set-chainlink-vrf-fee")
    add_default_arguments(set_chainlink_vrf_fee_parser, True)
    set_chainlink_vrf_fee_parser.add_argument(
        "--fee-arg", required=True, help="Type: uint256", type=int
    )
    set_chainlink_vrf_fee_parser.set_defaults(func=handle_set_chainlink_vrf_fee)

    set_chainlink_vrf_keyhash_parser = subcommands.add_parser(
        "set-chainlink-vrf-keyhash"
    )
    add_default_arguments(set_chainlink_vrf_keyhash_parser, True)
    set_chainlink_vrf_keyhash_parser.add_argument(
        "--keyhash-arg", required=True, help="Type: bytes32", type=bytes_argument_type
    )
    set_chainlink_vrf_keyhash_parser.set_defaults(func=handle_set_chainlink_vrf_keyhash)

    set_crypto_unicorns_address_parser = subcommands.add_parser(
        "set-crypto-unicorns-address"
    )
    add_default_arguments(set_crypto_unicorns_address_parser, True)
    set_crypto_unicorns_address_parser.add_argument(
        "--crypto-unicorns-address-arg", required=True, help="Type: address"
    )
    set_crypto_unicorns_address_parser.set_defaults(
        func=handle_set_crypto_unicorns_address
    )

    set_lootbox_pool_ids_parser = subcommands.add_parser("set-lootbox-pool-ids")
    add_default_arguments(set_lootbox_pool_ids_parser, True)
    set_lootbox_pool_ids_parser.add_argument(
        "--lootbox-pool-ids-arg", required=True, help="Type: uint256[3]", nargs="+"
    )
    set_lootbox_pool_ids_parser.set_defaults(func=handle_set_lootbox_pool_ids)

    set_shadowcorn_egg_bonuses_to_unicorn_milk_parser = subcommands.add_parser(
        "set-shadowcorn-egg-bonuses-to-unicorn-milk"
    )
    add_default_arguments(set_shadowcorn_egg_bonuses_to_unicorn_milk_parser, True)
    set_shadowcorn_egg_bonuses_to_unicorn_milk_parser.add_argument(
        "--bonuses-arg", required=True, help="Type: uint256[3]", nargs="+"
    )
    set_shadowcorn_egg_bonuses_to_unicorn_milk_parser.set_defaults(
        func=handle_set_shadowcorn_egg_bonuses_to_unicorn_milk
    )

    set_shadowcorn_pool_ids_parser = subcommands.add_parser("set-shadowcorn-pool-ids")
    add_default_arguments(set_shadowcorn_pool_ids_parser, True)
    set_shadowcorn_pool_ids_parser.add_argument(
        "--shadowcorn-pool-ids-arg", required=True, help="Type: uint256[3]", nargs="+"
    )
    set_shadowcorn_pool_ids_parser.set_defaults(func=handle_set_shadowcorn_pool_ids)

    set_stake_period_seconds_parser = subcommands.add_parser("set-stake-period-seconds")
    add_default_arguments(set_stake_period_seconds_parser, True)
    set_stake_period_seconds_parser.add_argument(
        "--stake-period-seconds-arg", required=True, help="Type: uint256", type=int
    )
    set_stake_period_seconds_parser.set_defaults(func=handle_set_stake_period_seconds)

    set_terminus_address_parser = subcommands.add_parser("set-terminus-address")
    add_default_arguments(set_terminus_address_parser, True)
    set_terminus_address_parser.add_argument(
        "--terminus-address-arg", required=True, help="Type: address"
    )
    set_terminus_address_parser.set_defaults(func=handle_set_terminus_address)

    set_unicorn_milk_staking_reward_parser = subcommands.add_parser(
        "set-unicorn-milk-staking-reward"
    )
    add_default_arguments(set_unicorn_milk_staking_reward_parser, True)
    set_unicorn_milk_staking_reward_parser.add_argument(
        "--unicorn-milk-staking-reward-arg",
        required=True,
        help="Type: uint256",
        type=int,
    )
    set_unicorn_milk_staking_reward_parser.set_defaults(
        func=handle_set_unicorn_milk_staking_reward
    )

    set_unicorn_milk_thresholds_parser = subcommands.add_parser(
        "set-unicorn-milk-thresholds"
    )
    add_default_arguments(set_unicorn_milk_thresholds_parser, True)
    set_unicorn_milk_thresholds_parser.add_argument(
        "--unicorn-milk-thresholds-arg",
        required=True,
        help="Type: uint256[3]",
        nargs="+",
    )
    set_unicorn_milk_thresholds_parser.set_defaults(
        func=handle_set_unicorn_milk_thresholds
    )

    set_unicorn_milk_token_address_parser = subcommands.add_parser(
        "set-unicorn-milk-token-address"
    )
    add_default_arguments(set_unicorn_milk_token_address_parser, True)
    set_unicorn_milk_token_address_parser.add_argument(
        "--unicorn-milk-token-address-arg", required=True, help="Type: address"
    )
    set_unicorn_milk_token_address_parser.set_defaults(
        func=handle_set_unicorn_milk_token_address
    )

    shadowcorn_egg_bonuses_to_unicorn_milk_parser = subcommands.add_parser(
        "shadowcorn-egg-bonuses-to-unicorn-milk"
    )
    add_default_arguments(shadowcorn_egg_bonuses_to_unicorn_milk_parser, False)
    shadowcorn_egg_bonuses_to_unicorn_milk_parser.set_defaults(
        func=handle_shadowcorn_egg_bonuses_to_unicorn_milk
    )

    shadowcorn_pool_ids_parser = subcommands.add_parser("shadowcorn-pool-ids")
    add_default_arguments(shadowcorn_pool_ids_parser, False)
    shadowcorn_pool_ids_parser.set_defaults(func=handle_shadowcorn_pool_ids)

    stake_period_seconds_parser = subcommands.add_parser("stake-period-seconds")
    add_default_arguments(stake_period_seconds_parser, False)
    stake_period_seconds_parser.set_defaults(func=handle_stake_period_seconds)

    staker_parser = subcommands.add_parser("staker")
    add_default_arguments(staker_parser, False)
    staker_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    staker_parser.set_defaults(func=handle_staker)

    surrender_terminus_pools_parser = subcommands.add_parser("surrender-terminus-pools")
    add_default_arguments(surrender_terminus_pools_parser, True)
    surrender_terminus_pools_parser.set_defaults(func=handle_surrender_terminus_pools)

    terminus_address_parser = subcommands.add_parser("terminus-address")
    add_default_arguments(terminus_address_parser, False)
    terminus_address_parser.set_defaults(func=handle_terminus_address)

    token_of_staker_by_index_parser = subcommands.add_parser("token-of-staker-by-index")
    add_default_arguments(token_of_staker_by_index_parser, False)
    token_of_staker_by_index_parser.add_argument(
        "--staker-arg", required=True, help="Type: address"
    )
    token_of_staker_by_index_parser.add_argument(
        "--index", required=True, help="Type: uint256", type=int
    )
    token_of_staker_by_index_parser.set_defaults(func=handle_token_of_staker_by_index)

    transfer_ownership_parser = subcommands.add_parser("transfer-ownership")
    add_default_arguments(transfer_ownership_parser, True)
    transfer_ownership_parser.add_argument(
        "--new-owner", required=True, help="Type: address"
    )
    transfer_ownership_parser.set_defaults(func=handle_transfer_ownership)

    unicorn_milk_staking_reward_parser = subcommands.add_parser(
        "unicorn-milk-staking-reward"
    )
    add_default_arguments(unicorn_milk_staking_reward_parser, False)
    unicorn_milk_staking_reward_parser.set_defaults(
        func=handle_unicorn_milk_staking_reward
    )

    unicorn_milk_thresholds_parser = subcommands.add_parser("unicorn-milk-thresholds")
    add_default_arguments(unicorn_milk_thresholds_parser, False)
    unicorn_milk_thresholds_parser.set_defaults(func=handle_unicorn_milk_thresholds)

    unicorn_milk_token_address_parser = subcommands.add_parser(
        "unicorn-milk-token-address"
    )
    add_default_arguments(unicorn_milk_token_address_parser, False)
    unicorn_milk_token_address_parser.set_defaults(
        func=handle_unicorn_milk_token_address
    )

    unstakes_at_parser = subcommands.add_parser("unstakes-at")
    add_default_arguments(unstakes_at_parser, False)
    unstakes_at_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    unstakes_at_parser.set_defaults(func=handle_unstakes_at)

    view_sacrifice_parser = subcommands.add_parser("view-sacrifice")
    add_default_arguments(view_sacrifice_parser, False)
    view_sacrifice_parser.add_argument(
        "--request-id-arg",
        required=True,
        help="Type: bytes32",
        type=bytes_argument_type,
    )
    view_sacrifice_parser.set_defaults(func=handle_view_sacrifice)

    withdraw_erc20_parser = subcommands.add_parser("withdraw-erc20")
    add_default_arguments(withdraw_erc20_parser, True)
    withdraw_erc20_parser.add_argument(
        "--token-address", required=True, help="Type: address"
    )
    withdraw_erc20_parser.add_argument(
        "--amount", required=True, help="Type: uint256", type=int
    )
    withdraw_erc20_parser.set_defaults(func=handle_withdraw_erc20)

    return parser


def main() -> None:
    parser = generate_cli()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# Code generated by moonworm : https://github.com/bugout-dev/moonworm
# Moonworm version : 0.1.15

import argparse
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from brownie import Contract, network, project
from brownie.network.contract import ContractContainer
from eth_typing.evm import ChecksumAddress


PROJECT_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
BUILD_DIRECTORY = os.path.join(PROJECT_DIRECTORY, "build", "contracts")


def boolean_argument_type(raw_value: str) -> bool:
    TRUE_VALUES = ["1", "t", "y", "true", "yes"]
    FALSE_VALUES = ["0", "f", "n", "false", "no"]

    if raw_value.lower() in TRUE_VALUES:
        return True
    elif raw_value.lower() in FALSE_VALUES:
        return False

    raise ValueError(
        f"Invalid boolean argument: {raw_value}. Value must be one of: {','.join(TRUE_VALUES + FALSE_VALUES)}"
    )


def bytes_argument_type(raw_value: str) -> bytes:
    return raw_value.encode()


def get_abi_json(abi_name: str) -> List[Dict[str, Any]]:
    abi_full_path = os.path.join(BUILD_DIRECTORY, f"{abi_name}.json")
    if not os.path.isfile(abi_full_path):
        raise IOError(
            f"File does not exist: {abi_full_path}. Maybe you have to compile the smart contracts?"
        )

    with open(abi_full_path, "r") as ifp:
        build = json.load(ifp)

    abi_json = build.get("abi")
    if abi_json is None:
        raise ValueError(f"Could not find ABI definition in: {abi_full_path}")

    return abi_json


def contract_from_build(abi_name: str) -> ContractContainer:
    # This is workaround because brownie currently doesn't support loading the same project multiple
    # times. This causes problems when using multiple contracts from the same project in the same
    # python project.
    PROJECT = project.main.Project("moonworm", Path(PROJECT_DIRECTORY))

    abi_full_path = os.path.join(BUILD_DIRECTORY, f"{abi_name}.json")
    if not os.path.isfile(abi_full_path):
        raise IOError(
            f"File does not exist: {abi_full_path}. Maybe you have to compile the smart contracts?"
        )

    with open(abi_full_path, "r") as ifp:
        build = json.load(ifp)

    return ContractContainer(PROJECT, build)


class ERC721:
    def __init__(self, contract_address: Optional[ChecksumAddress]):
        self.contract_name = "ERC721"
        self.address = contract_address
        self.contract = None
        self.abi = get_abi_json("ERC721")
        if self.address is not None:
            self.contract: Optional[Contract] = Contract.from_abi(
                self.contract_name, self.address, self.abi
            )

    def deploy(
        self, name_: str, symbol_: str, owner: ChecksumAddress, transaction_config
    ):
        contract_class = contract_from_build(self.contract_name)
        deployed_contract = contract_class.deploy(
            name_, symbol_, owner, transaction_config
        )
        self.address = deployed_contract.address
        self.contract = deployed_contract

    def assert_contract_is_instantiated(self) -> None:
        if self.contract is None:
            raise Exception("contract has not been instantiated")

    def verify_contract(self):
        self.assert_contract_is_instantiated()
        contract_class = contract_from_build(self.contract_name)
        contract_class.publish_source(self.contract)

    def approve(self, to: ChecksumAddress, token_id: int, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.approve(to, token_id, transaction_config)

    def balance_of(self, owner: ChecksumAddress) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.balanceOf.call(owner)

    def get_approved(self, token_id: int) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.getApproved.call(token_id)

    def is_approved_for_all(
        self, owner: ChecksumAddress, operator: ChecksumAddress
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.isApprovedForAll.call(owner, operator)

    def mint(self, to: ChecksumAddress, token_id: int, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.mint(to, token_id, transaction_config)

    def name(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.name.call()

    def owner(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.owner.call()

    def owner_of(self, token_id: int) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.ownerOf.call(token_id)

    def renounce_ownership(self, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.renounceOwnership(transaction_config)

    def safe_transfer_from(
        self,
        from_: ChecksumAddress,
        to: ChecksumAddress,
        token_id: int,
        transaction_config,
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.safeTransferFrom(from_, to, token_id, transaction_config)

    def safe_transfer_from(
        self,
        from_: ChecksumAddress,
        to: ChecksumAddress,
        token_id: int,
        _data: bytes,
        transaction_config,
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.safeTransferFrom(
            from_, to, token_id, _data, transaction_config
        )

    def set_approval_for_all(
        self, operator: ChecksumAddress, approved: bool, transaction_config
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.setApprovalForAll(operator, approved, transaction_config)

    def supports_interface(self, interface_id: bytes) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.supportsInterface.call(interface_id)

    def symbol(self) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.symbol.call()

    def token_uri(self, token_id: int) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.tokenURI.call(token_id)

    def transfer_from(
        self,
        from_: ChecksumAddress,
        to: ChecksumAddress,
        token_id: int,
        transaction_config,
    ) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.transferFrom(from_, to, token_id, transaction_config)

    def transfer_ownership(self, new_owner: ChecksumAddress, transaction_config) -> Any:
        self.assert_contract_is_instantiated()
        return self.contract.transferOwnership(new_owner, transaction_config)


def get_transaction_config(args: argparse.Namespace) -> Dict[str, Any]:
    signer = network.accounts.load(args.sender, args.password)
    transaction_config: Dict[str, Any] = {"from": signer}
    if args.gas_price is not None:
        transaction_config["gas_price"] = args.gas_price
    if args.max_fee_per_gas is not None:
        transaction_config["max_fee"] = args.max_fee_per_gas
    if args.max_priority_fee_per_gas is not None:
        transaction_config["priority_fee"] = args.max_priority_fee_per_gas
    if args.confirmations is not None:
        transaction_config["required_confs"] = args.confirmations
    if args.nonce is not None:
        transaction_config["nonce"] = args.nonce
    return transaction_config


def add_default_arguments(parser: argparse.ArgumentParser, transact: bool) -> None:
    parser.add_argument(
        "--network", required=True, help="Name of brownie network to connect to"
    )
    parser.add_argument(
        "--address", required=False, help="Address of deployed contract to connect to"
    )
    if not transact:
        return
    parser.add_argument(
        "--sender", required=True, help="Path to keystore file for transaction sender"
    )
    parser.add_argument(
        "--password",
        required=False,
        help="Password to keystore file (if you do not provide it, you will be prompted for it)",
    )
    parser.add_argument(
        "--gas-price", default=None, help="Gas price at which to submit transaction"
    )
    parser.add_argument(
        "--max-fee-per-gas",
        default=None,
        help="Max fee per gas for EIP1559 transactions",
    )
    parser.add_argument(
        "--max-priority-fee-per-gas",
        default=None,
        help="Max priority fee per gas for EIP1559 transactions",
    )
    parser.add_argument(
        "--confirmations",
        type=int,
        default=None,
        help="Number of confirmations to await before considering a transaction completed",
    )
    parser.add_argument(
        "--nonce", type=int, default=None, help="Nonce for the transaction (optional)"
    )
    parser.add_argument(
        "--value", default=None, help="Value of the transaction in wei(optional)"
    )


def handle_deploy(args: argparse.Namespace) -> None:
    network.connect(args.network)
    transaction_config = get_transaction_config(args)
    contract = ERC721(None)
    result = contract.deploy(
        name_=args.name_arg,
        symbol_=args.symbol_arg,
        owner=args.owner,
        transaction_config=transaction_config,
    )
    print(result)


def handle_verify_contract(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.verify_contract()
    print(result)


def handle_approve(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.approve(
        to=args.to, token_id=args.token_id, transaction_config=transaction_config
    )
    print(result)


def handle_balance_of(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.balance_of(owner=args.owner)
    print(result)


def handle_get_approved(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.get_approved(token_id=args.token_id)
    print(result)


def handle_is_approved_for_all(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.is_approved_for_all(owner=args.owner, operator=args.operator)
    print(result)


def handle_mint(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.mint(
        to=args.to, token_id=args.token_id, transaction_config=transaction_config
    )
    print(result)


def handle_name(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.name()
    print(result)


def handle_owner(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.owner()
    print(result)


def handle_owner_of(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.owner_of(token_id=args.token_id)
    print(result)


def handle_renounce_ownership(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.renounce_ownership(transaction_config=transaction_config)
    print(result)


def handle_safe_transfer_from(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.safe_transfer_from(
        from_=args.from_arg,
        to=args.to,
        token_id=args.token_id,
        transaction_config=transaction_config,
    )
    print(result)


def handle_safe_transfer_from(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.safe_transfer_from(
        from_=args.from_arg,
        to=args.to,
        token_id=args.token_id,
        _data=args.data_arg,
        transaction_config=transaction_config,
    )
    print(result)


def handle_set_approval_for_all(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.set_approval_for_all(
        operator=args.operator,
        approved=args.approved,
        transaction_config=transaction_config,
    )
    print(result)


def handle_supports_interface(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.supports_interface(interface_id=args.interface_id)
    print(result)


def handle_symbol(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.symbol()
    print(result)


def handle_token_uri(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    result = contract.token_uri(token_id=args.token_id)
    print(result)


def handle_transfer_from(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.transfer_from(
        from_=args.from_arg,
        to=args.to,
        token_id=args.token_id,
        transaction_config=transaction_config,
    )
    print(result)


def handle_transfer_ownership(args: argparse.Namespace) -> None:
    network.connect(args.network)
    contract = ERC721(args.address)
    transaction_config = get_transaction_config(args)
    result = contract.transfer_ownership(
        new_owner=args.new_owner, transaction_config=transaction_config
    )
    print(result)


def generate_cli() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="CLI for ERC721")
    parser.set_defaults(func=lambda _: parser.print_help())
    subcommands = parser.add_subparsers()

    deploy_parser = subcommands.add_parser("deploy")
    add_default_arguments(deploy_parser, True)
    deploy_parser.add_argument(
        "--name-arg", required=True, help="Type: string", type=str
    )
    deploy_parser.add_argument(
        "--symbol-arg", required=True, help="Type: string", type=str
    )
    deploy_parser.add_argument("--owner", required=True, help="Type: address")
    deploy_parser.set_defaults(func=handle_deploy)

    verify_contract_parser = subcommands.add_parser("verify-contract")
    add_default_arguments(verify_contract_parser, False)
    verify_contract_parser.set_defaults(func=handle_verify_contract)

    approve_parser = subcommands.add_parser("approve")
    add_default_arguments(approve_parser, True)
    approve_parser.add_argument("--to", required=True, help="Type: address")
    approve_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    approve_parser.set_defaults(func=handle_approve)

    balance_of_parser = subcommands.add_parser("balance-of")
    add_default_arguments(balance_of_parser, False)
    balance_of_parser.add_argument("--owner", required=True, help="Type: address")
    balance_of_parser.set_defaults(func=handle_balance_of)

    get_approved_parser = subcommands.add_parser("get-approved")
    add_default_arguments(get_approved_parser, False)
    get_approved_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    get_approved_parser.set_defaults(func=handle_get_approved)

    is_approved_for_all_parser = subcommands.add_parser("is-approved-for-all")
    add_default_arguments(is_approved_for_all_parser, False)
    is_approved_for_all_parser.add_argument(
        "--owner", required=True, help="Type: address"
    )
    is_approved_for_all_parser.add_argument(
        "--operator", required=True, help="Type: address"
    )
    is_approved_for_all_parser.set_defaults(func=handle_is_approved_for_all)

    mint_parser = subcommands.add_parser("mint")
    add_default_arguments(mint_parser, True)
    mint_parser.add_argument("--to", required=True, help="Type: address")
    mint_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    mint_parser.set_defaults(func=handle_mint)

    name_parser = subcommands.add_parser("name")
    add_default_arguments(name_parser, False)
    name_parser.set_defaults(func=handle_name)

    owner_parser = subcommands.add_parser("owner")
    add_default_arguments(owner_parser, False)
    owner_parser.set_defaults(func=handle_owner)

    owner_of_parser = subcommands.add_parser("owner-of")
    add_default_arguments(owner_of_parser, False)
    owner_of_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    owner_of_parser.set_defaults(func=handle_owner_of)

    renounce_ownership_parser = subcommands.add_parser("renounce-ownership")
    add_default_arguments(renounce_ownership_parser, True)
    renounce_ownership_parser.set_defaults(func=handle_renounce_ownership)

    safe_transfer_from_parser = subcommands.add_parser("safe-transfer-from")
    add_default_arguments(safe_transfer_from_parser, True)
    safe_transfer_from_parser.add_argument(
        "--from-arg", required=True, help="Type: address"
    )
    safe_transfer_from_parser.add_argument("--to", required=True, help="Type: address")
    safe_transfer_from_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    safe_transfer_from_parser.set_defaults(func=handle_safe_transfer_from)

    safe_transfer_from_parser = subcommands.add_parser("safe-transfer-from")
    add_default_arguments(safe_transfer_from_parser, True)
    safe_transfer_from_parser.add_argument(
        "--from-arg", required=True, help="Type: address"
    )
    safe_transfer_from_parser.add_argument("--to", required=True, help="Type: address")
    safe_transfer_from_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    safe_transfer_from_parser.add_argument(
        "--data-arg", required=True, help="Type: bytes", type=bytes_argument_type
    )
    safe_transfer_from_parser.set_defaults(func=handle_safe_transfer_from)

    set_approval_for_all_parser = subcommands.add_parser("set-approval-for-all")
    add_default_arguments(set_approval_for_all_parser, True)
    set_approval_for_all_parser.add_argument(
        "--operator", required=True, help="Type: address"
    )
    set_approval_for_all_parser.add_argument(
        "--approved", required=True, help="Type: bool", type=boolean_argument_type
    )
    set_approval_for_all_parser.set_defaults(func=handle_set_approval_for_all)

    supports_interface_parser = subcommands.add_parser("supports-interface")
    add_default_arguments(supports_interface_parser, False)
    supports_interface_parser.add_argument(
        "--interface-id", required=True, help="Type: bytes4", type=bytes_argument_type
    )
    supports_interface_parser.set_defaults(func=handle_supports_interface)

    symbol_parser = subcommands.add_parser("symbol")
    add_default_arguments(symbol_parser, False)
    symbol_parser.set_defaults(func=handle_symbol)

    token_uri_parser = subcommands.add_parser("token-uri")
    add_default_arguments(token_uri_parser, False)
    token_uri_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    token_uri_parser.set_defaults(func=handle_token_uri)

    transfer_from_parser = subcommands.add_parser("transfer-from")
    add_default_arguments(transfer_from_parser, True)
    transfer_from_parser.add_argument("--from-arg", required=True, help="Type: address")
    transfer_from_parser.add_argument("--to", required=True, help="Type: address")
    transfer_from_parser.add_argument(
        "--token-id", required=True, help="Type: uint256", type=int
    )
    transfer_from_parser.set_defaults(func=handle_transfer_from)

    transfer_ownership_parser = subcommands.add_parser("transfer-ownership")
    add_default_arguments(transfer_ownership_parser, True)
    transfer_ownership_parser.add_argument(
        "--new-owner", required=True, help="Type: address"
    )
    transfer_ownership_parser.set_defaults(func=handle_transfer_ownership)

    return parser


def main() -> None:
    parser = generate_cli()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



from setuptools import find_packages, setup

long_description = ""
with open("README.md") as ifp:
    long_description = ifp.read()

setup(
    name="autocorns",
    version="0.0.5",
    packages=find_packages(),
    package_data={"autocorns": ["build/contracts/*.json"]},
    include_package_data=True,
    install_requires=["eth-brownie", "tqdm", "web3"],
    extras_require={
        "dev": [
            "black",
            "moonworm >= 0.1.14",
        ],
        "distribute": ["setuptools", "twine", "wheel"],
    },
    description="A team of Crypto Unicorn bots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="zomglings",
    author_email="nkashy1@gmail.com",
    url="https://github.com/zomglings/autocorns",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
    ],
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "autocorns=autocorns.cli:main",
        ]
    },
)



"""
The warden leads unicorns safely through the Dark Forest.
"""

import argparse
import json
import os
import time
from typing import List, Tuple

from brownie import network

from . import DarkForest, ERC721

CU_MAINNET_ADDRESS = "0xdC0479CC5BbA033B3e7De9F178607150B3AbCe1f"
DARK_FOREST_MAINNET_ADDRESS = "0x8d528e98A69FE27b11bb02Ac264516c4818C3942"

BROWNIE_NETWORK = os.environ.get("BROWNIE_NETWORK", "matic")
DARK_FOREST_ADDRESS = os.environ.get("DARK_FOREST_ADDRESS", DARK_FOREST_MAINNET_ADDRESS)
CU_ADDRESS = os.environ.get("DARK_FOREST_ADDRESS", CU_MAINNET_ADDRESS)


def escort(corns: List[int], transaction_config) -> List[int]:
    """
    Unstakes all unstakable unicorns for the given player.
    Then stakes all stakable unicorns for the given player.

    Return list of staked corns.
    """
    print(
        f"Network: {BROWNIE_NETWORK}, Crypto Unicorns: {CU_ADDRESS}, Dark Forest: {DARK_FOREST_ADDRESS}"
    )
    network.connect(BROWNIE_NETWORK)

    nonce = None
    if transaction_config.get("nonce"):
        nonce = transaction_config["nonce"] - 1

    unstaked: List[int] = []
    staked: List[Tuple[int, int]] = []

    crypto_unicorns = ERC721.ERC721(CU_ADDRESS)
    dark_forest = DarkForest.DarkForest(DARK_FOREST_ADDRESS)

    for corn in corns:
        unstakes_at = dark_forest.unstakes_at(corn)
        if unstakes_at > 0:
            staked.append((corn, unstakes_at))
        else:
            unstaked.append(corn)

    time_now = int(time.time())
    print(f"Machine time: {time_now}")
    staked_final: List[int] = []
    ready_to_unstake: List[int] = []

    for corn, unstakes_at in staked:
        if unstakes_at <= time_now:
            ready_to_unstake.append(corn)
        else:
            staked_final.append(corn)

    for corn in ready_to_unstake:
        if nonce is not None:
            nonce += 1
            transaction_config["nonce"] = nonce
        dark_forest.exit_forest(corn, transaction_config)
        unstaked.append(corn)

    player = transaction_config["from"].address
    for corn in unstaked:
        if nonce is not None:
            nonce += 1
            transaction_config["nonce"] = nonce
        crypto_unicorns.safe_transfer_from(
            player, DARK_FOREST_ADDRESS, corn, b"", transaction_config
        )
        staked_final.append(corn)

    return staked_final


def handle_escort(args: argparse.Namespace) -> None:
    transaction_config = DarkForest.get_transaction_config(args)
    staked_corns = escort(args.corns, transaction_config)
    print(json.dumps(staked_corns))


def generate_cli() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="The Dark Forest warden")

    DarkForest.add_default_arguments(parser, True)
    parser.add_argument(
        "--corns",
        nargs="+",
        type=int,
        help="List of corns to escort into or out of the Dark Forest",
    )

    parser.set_defaults(func=handle_escort)

    return parser


if __name__ == "__main__":
    parser = generate_cli()
    args = parser.parse_args()
    args.func(args)



import argparse

from . import warden, biologist


def main():
    parser = argparse.ArgumentParser(
        description="autocorns: Crypto Unicorns automation"
    )
    parser.set_defaults(func=lambda _: parser.print_help())
    subparsers = parser.add_subparsers()

    warden_parser = warden.generate_cli()
    subparsers.add_parser("warden", parents=[warden_parser], add_help=False)

    biologist_parser = biologist.generate_cli()
    subparsers.add_parser("biologist", parents=[biologist_parser], add_help=False)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()
